/*
	Connection.cpp
*/

#include "stdafx.h"
#include "blobfn.h"
#include "unaccfn.h"
#include "LiteX.h"
#include "Connection.h"
#include "Statement.h"
#include "Glob.h"
#include "Messages.h"

STDMETHODIMP CLiteConnection::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_ILiteConnection
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (IsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

HRESULT CLiteConnection::sqlite_error( int nError ) const
{
	return litex_error( sqlite_2_hr( nError ) );
}

HRESULT CLiteConnection::litex_error( HRESULT hr ) const
{
	CComBSTR bsDesc;
	if ( FAILED(hr) && get_description( hr, bsDesc ) )
	{
		return Error( bsDesc, IID_ILiteConnection, hr );
	}
	else
	{	
		return hr;
	}
}

HRESULT CLiteConnection::sqlite_error( int nError, const char* pszMsg ) const
{
	HRESULT hr = sqlite_2_hr( nError );
	if ( FAILED(hr) )
	{
		return Error( pszMsg, IID_ILiteConnection, hr );
	}
	else
	{
		return hr;
	}
}

STDMETHODIMP CLiteConnection::GetDB( void** pDB )
{
	*pDB = m_db;
	return S_OK;
}

STDMETHODIMP CLiteConnection::get_Version( VARIANT_BOOL bString, VARIANT *pVal)
{
	CComVariant v;
	
	if ( VARIANT_FALSE == bString )
	{ // numerical version
		v = sqlite3_libversion_number();
	}
	else
	{ // string version
		CComBSTR bsVersion( sqlite3_libversion() );	
		v = bsVersion;
	}
	return v.Detach( pVal );
}

STDMETHODIMP CLiteConnection::get_Path(BSTR *pVal)
{
	m_bsPath.CopyTo( pVal );
	return S_OK;
}

STDMETHODIMP CLiteConnection::put_Path(BSTR newVal)
{
	if ( m_db != NULL )
	{
		return litex_error( E_DB_IS_OPEN );
	}

	m_bsPath = newVal;
	return S_OK;
}

STDMETHODIMP CLiteConnection::get_Changes( LONG *pnVal )
{
	if ( m_db == NULL )
	{
		return litex_error( E_DB_IS_NOT_OPEN );
	}

	*pnVal = sqlite3_changes( m_db );
	return S_OK;
}

STDMETHODIMP CLiteConnection::get_LastInsertRowid( VARIANT *pVal )
{
	if ( m_db == NULL )
	{
		return litex_error( E_DB_IS_NOT_OPEN );
	}

	CComVariant v;
	HRESULT hr = make_integer_or_decimal( v, sqlite3_last_insert_rowid( m_db ) );
	if ( FAILED(hr) )
	{
		return litex_error( hr );
	}
	return v.Detach( pVal );
}

STDMETHODIMP CLiteConnection::Open(BSTR bsPath)
{
#ifndef _UNICODE
	USES_CONV_UTF8;
#endif

	if ( m_db != NULL )
	{
		return litex_error( E_DB_IS_OPEN );
	}

	if ( SysStringLen( bsPath ) > 0 )
	{
		put_Path( bsPath );
	}

	int nRes;
#ifdef _UNICODE
	nRes = sqlite3_open16( m_bsPath, &m_db );
#else
	nRes = sqlite3_open( W2UTF8( m_bsPath ), &m_db );
#endif

	if ( SQLITE_OK == nRes )
	{
		sqlite3_register_blob_functions( m_db );
#ifndef _NO_UNACC_STUFF
		m_vUnaccContext = sqlite3_register_unacc_functions( m_db );
#endif
		sqlite3_busy_handler( m_db, &busy_callback, this );
		sqlite3_progress_handler( m_db, m_nProgressPeriod, &progress_callback, this );
	}

	return sqlite_error( nRes );
}

STDMETHODIMP CLiteConnection::OpenInMemory()
{
	CComBSTR bsPath( _T(":memory:") );
	return Open( bsPath );
}

STDMETHODIMP CLiteConnection::Close()
{
	int nRes = SQLITE_OK;

	if ( m_db )
	{
		sqlite3_interrupt( m_db );
#ifndef _NO_UNACC_STUFF
		sqlite3_unregister_unacc_functions( m_db, m_vUnaccContext );
#endif
		sqlite3_busy_handler( m_db, NULL, 0 );
		sqlite3_progress_handler( m_db, DEF_PROGRESS_PERIOD, NULL, 0 );
		
		nRes = sqlite3_close( m_db );
		if ( nRes == SQLITE_OK ) m_db = NULL;
	}

	return sqlite_error( nRes );
}

HRESULT CLiteConnection::create_statement( CComPtr<ILiteStatement>& pStmt )
{
	CComObject< CLiteStatement >* p = new CComObject< CLiteStatement >;
	HRESULT hr = p->FinalConstruct();
	if ( FAILED(hr) )
	{
		delete p;
		return hr;
	}

	hr = p->QueryInterface( &pStmt );
	if ( FAILED(hr) )
	{
		delete p;
		return hr;
	}

	CComPtr<IDispatch> pThis;
	hr = QueryInterface( IID_IDispatch, (void**)&pThis );
	if ( FAILED(hr) ) return hr;

	hr = pStmt->put_ActiveConnection( pThis );
	if ( FAILED(hr) ) return hr;

	return S_OK;
}

STDMETHODIMP CLiteConnection::Prepare(BSTR bsCommand, IDispatch **ppStmt)
{
	if ( m_db == NULL )
	{
		return litex_error( E_DB_IS_NOT_OPEN );
	}

	CComPtr<ILiteStatement> pStmt;
	HRESULT hr = create_statement( pStmt );
	if ( FAILED(hr) )
	{
		return litex_error( E_CANNOT_CREATE_STATEMENT );
	}

	hr = pStmt->Prepare( bsCommand );
	if ( FAILED(hr) ) return hr;

	return pStmt->QueryInterface( ppStmt );
}

STDMETHODIMP CLiteConnection::Execute(BSTR bsCommand, LPSAFEARRAY *params, VARIANT *pVal )
{
	if ( m_db == NULL )
	{
		return litex_error( E_DB_IS_NOT_OPEN );
	}

	CComPtr<ILiteStatement> pStmt;
	HRESULT hr = create_statement( pStmt );
	if ( FAILED(hr) )
	{
		return litex_error( E_CANNOT_CREATE_STATEMENT );
	}

	hr = pStmt->Prepare( bsCommand );
	if ( FAILED(hr) ) return hr;
	
	hr = pStmt->BindParameters( params );
	if ( FAILED(hr) ) return hr;	

	VARIANT_BOOL bDone;
	hr = pStmt->Step( 1, &bDone );
	if ( FAILED(hr) ) return hr;

	if ( bDone == VARIANT_FALSE )
	{
		hr = pStmt->get_Row( lxDefault, pVal );
		if ( FAILED(hr) ) return hr;
	}
	else
	{
		pVal->vt = VT_EMPTY;
	}

	hr = pStmt->Close();
	if ( FAILED(hr) ) return hr;

	return S_OK;
}

STDMETHODIMP CLiteConnection::BatchExecute(BSTR bsSql)
{
	if ( m_db == NULL )
	{
		return litex_error( E_DB_IS_NOT_OPEN );
	}
	
	USES_CONV_UTF8;
	
	char* pszErrMsg = NULL;
	int nRes = sqlite3_exec( m_db, W2UTF8(bsSql), NULL, NULL, &pszErrMsg );
	HRESULT hr = sqlite_error( nRes, pszErrMsg );
	sqlite3_free( pszErrMsg );
	return hr;
}

int CLiteConnection::busy_callback( void* _pThis, int nTimes )
{
	CLiteConnection* pThis = static_cast<CLiteConnection*>( _pThis );
	VARIANT_BOOL bResult;
	HRESULT hRes = pThis->Fire_Busy( nTimes, &bResult );
	
	if ( SUCCEEDED(hRes) )
	{
		return (VARIANT_FALSE==bResult)? 0 : 1;
	}
	else
	{
		return 0;
	}
}

int CLiteConnection::progress_callback( void* _pThis )
{
	CLiteConnection* pThis = static_cast<CLiteConnection*>( _pThis );
	VARIANT_BOOL bResult;
	HRESULT hRes = pThis->Fire_Progress( &bResult );
	if ( SUCCEEDED(hRes) )
	{
		return (VARIANT_FALSE==bResult)? 1 : 0;
	}
	else
	{
		return 0;
	}
}

STDMETHODIMP CLiteConnection::get_ProgressPeriod(LONG *pnPeriod)
{
	*pnPeriod = m_nProgressPeriod;
	return S_OK;
}

STDMETHODIMP CLiteConnection::put_ProgressPeriod(LONG nPeriod)
{
	m_nProgressPeriod = nPeriod;
	if ( m_db != NULL )
	{
		sqlite3_progress_handler( m_db, m_nProgressPeriod, progress_callback, this );
	}
	return S_OK;
}

STDMETHODIMP CLiteConnection::Interrupt()
{
	if ( m_db == NULL )
	{
		return litex_error( E_DB_IS_NOT_OPEN );
	}
	sqlite3_interrupt( m_db );
	return S_OK;
}

