/*
	Parameters.cpp
*/

#include "stdafx.h"
#include "Parameters.h"
#include "Parameter.h"
#include "Glob.h"
#include "Messages.h"

STDMETHODIMP CLiteParameters::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_ILiteParameters
	};

	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (IsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

HRESULT CLiteParameters::litex_error( HRESULT hr ) const
{
	CComBSTR bsDesc;
	if ( FAILED(hr) && get_description( hr, bsDesc ) )
	{
		return Error( bsDesc, IID_ILiteStatement, hr );
	}
	else
	{
		return hr;
	}
}

HRESULT CLiteParameters::Init( ILiteStatement* pStmt )
{
	HRESULT hRes;
	
	m_pStmt = pStmt;
	hRes = m_pStmt->get_ParameterCount( &m_nCount );
	if ( FAILED(hRes) )
	{
		return hRes;
	}
	
	SAFEARRAY* sa = SafeArrayCreateVectorEx( VT_DISPATCH, 0, m_nCount, (PVOID)&IID_IDispatch );
	if ( sa == NULL )
	{
		return E_FAIL;
	}
	
	for( LONG i=0; i<m_nCount; i++ )
	{
		CComPtr<IDispatch> pParameter;
		hRes = CLiteParameter::CreateObject( m_pStmt, i, &pParameter );
		if ( FAILED(hRes) )
		{
			break;
		}
		
		hRes = SafeArrayPutElement( sa, &i, pParameter );
		if ( FAILED(hRes) )
		{
			break;
		}
	}

	if ( FAILED(hRes) )
	{
		SafeArrayDestroy( sa );
		return hRes;
	}
	else
	{
		m_sa = sa;
	}
	
	return S_OK;	

}

HRESULT CLiteParameters::CreateObject( ILiteStatement* pStmt, IDispatch** ppObj )
{
	typedef CComObject<CLiteParameters> _class;
	HRESULT hRes = E_POINTER;
	if (ppObj != NULL)
	{
		*ppObj = NULL;
		_class* p;
		hRes = _class::CreateInstance(&p);
		if (SUCCEEDED(hRes))
		{
			// If this object has ownership of the data then we need to keep it around
			hRes = p->Init(pStmt);
			if (SUCCEEDED(hRes))
			{
				hRes = p->_InternalQueryInterface( IID_IDispatch, (void**)ppObj);
			}
			
			if (FAILED(hRes))
			{
				delete p;
			}
		}
	}
	return hRes;
}

STDMETHODIMP CLiteParameters::get__NewEnum(IUnknown** pVal)
{
	CComPtr<IUnknown> pThis;
	HRESULT hRes = _InternalQueryInterface( IID_IUnknown, (void**)&pThis );
	if ( FAILED(hRes) )
	{
		return hRes;
	}

	return CLiteParametersEnum::CreateObject( m_sa, pThis, pVal );
}

STDMETHODIMP CLiteParameters::get_Item(VARIANT vIndex, IDispatch** pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	
	HRESULT hRes = S_OK;
	CComVariant v( vIndex );
	hRes = variant_normalize( v );
	if ( FAILED(hRes) ) return litex_error( hRes );
	
	switch( V_VT(&v) & VT_TYPEMASK )
	{
		case VT_INT:
		{
			CComPtr<IDispatch> pParameter;		
			hRes = SafeArrayGetElement( m_sa, (LONG*)&V_INT(&v), &pParameter );
			if ( S_OK == hRes )
			{
				hRes = pParameter->QueryInterface( IID_IDispatch, (void**)pVal );
			}
		}
		break;
		
		case VT_BSTR:
		{
			CComBSTR bsName;
			bool bFind = false;
			for( LONG i=0; i<m_nCount; i++ )
			{
				CComPtr<IDispatch> pVParameter;
				hRes = SafeArrayGetElement( m_sa, &i, &pVParameter );
				if ( S_OK != hRes ) break;
				
				CComPtr<ILiteParameter> pParameter;
				hRes = pVParameter.QueryInterface( &pParameter );
				if ( S_OK != hRes ) break;
				
				hRes = pParameter->get_Name( &bsName );
				if ( S_OK != hRes ) break;
				
				if ( VARCMP_EQ == compare_bstr( V_BSTR(&v), bsName ) )
				{
					hRes = pParameter->QueryInterface( IID_IDispatch, (void**)pVal );
					bFind = true;
					break;
				}
			}
			
			if ( (S_OK==hRes) && !bFind )
			{
				hRes = E_UNKNOWN_PARAM_NAME;
			}
			
		}
		break;
		
		default:
		ATLASSERT( FALSE );
		break;
	}
	
	if ( FAILED(hRes) ) return litex_error( hRes );
	return hRes;
}

STDMETHODIMP CLiteParameters::get_Count(LONG* pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	*pVal = m_nCount;
	return S_OK;
}

STDMETHODIMP CLiteParameters::Bind(LPSAFEARRAY *params)
{
	if ( params == NULL )
	{
		return E_POINTER;
	}
	
	HRESULT hRes = m_pStmt->BindParameters( params );
	return hRes;
}
