/*
	Row.cpp
*/

#include "stdafx.h"
#include "Row.h"
#include "Glob.h"
#include "Messages.h"

STDMETHODIMP CLiteRow::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_ILiteRow
	};

	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (IsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

HRESULT CLiteRow::litex_error( HRESULT hr ) const
{
	CComBSTR bsDesc;
	if ( FAILED(hr) && get_description( hr, bsDesc ) )
	{
		return Error( bsDesc, IID_ILiteStatement, hr );
	}
	else
	{
		return hr;
	}
}



HRESULT CLiteRow::Init( ILiteStatement* _pStmt )
{
	CComPtr<ILiteStatement> pStmt( _pStmt );
	HRESULT hRes;
	
	hRes = pStmt->get_ColumnCount( &m_nSize );
	if ( FAILED(hRes) )
	{
		return hRes;
	}
	
	CComVariant var;
	hRes = pStmt->get_Row( lxArray, &var );
	if ( FAILED(hRes) )
	{
		return hRes;
	}
	
	ATLASSERT( V_VT(&var) == (VT_ARRAY|VT_VARIANT) );
	ATLASSERT( V_ARRAY(&var) != NULL );
	
	hRes = SafeArrayCopy( V_ARRAY(&var), &m_sa );
	if ( FAILED(hRes) )
	{
		return hRes;
	}

	return hRes;
}


STDMETHODIMP CLiteRow::get_Value(VARIANT* pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	
	HRESULT hRes = S_OK;
	LONG nIndex = 0;
	
	switch( m_nSize )
	{
		case 0:
		hRes = VariantClear( pVal );
		break;
		
		case 1:
		hRes = SafeArrayGetElement( m_sa, &nIndex, pVal );
		break;
		
		default:
		{
			CComVariant var;
			V_VT(&var) = VT_ARRAY | VT_VARIANT;
			hRes = SafeArrayCopy( m_sa, &var.parray );
			if ( SUCCEEDED(hRes) )
			{
				var.Detach( pVal );
			}
		}
		break;
	}

	return S_OK;
}

STDMETHODIMP CLiteRow::get__NewEnum(IUnknown** pVal)
{
	CComPtr<IUnknown> pThis;
	HRESULT hRes = _InternalQueryInterface( IID_IUnknown, (void**)&pThis );
	if ( FAILED(hRes) )
	{
		return hRes;
	}

	return CLiteRowEnum::CreateObject( m_sa, pThis, pVal );
}

STDMETHODIMP CLiteRow::get_Item(LONG nIndex, LxTypeEnum eType, VARIANT* pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	
	CComVariant val;
	HRESULT hRes = SafeArrayGetElement( m_sa, &nIndex, &val );
	if ( FAILED(hRes) )
	{
		return hRes;
	}
	
	switch( eType )
	{
		case lxUnknown: // nothing to do
		break;
		
		case lxInteger:
		case lxLongInteger:
		{
			CComVariant vint;
			hRes = vint.ChangeType( VT_I8, &val );
			if ( S_OK == hRes )
			{
				hRes = make_integer_or_decimal( val, V_I8(&vint) );
			}
		}
		break;
		
		case lxFloat:
		{
			hRes = val.ChangeType( VT_R8 );
		}
		break;
		
		case lxString:
		{
			hRes = val.ChangeType( VT_BSTR );
		}
		break;
		
		case lxDate:
		{
			CComVariant vdate;
			hRes = vdate.ChangeType( VT_R8, &val );
			if ( S_OK == hRes )
			{
				V_VT(&val) = VT_DATE;
				V_DATE(&val) = gat_automation_date( V_R8(&vdate) );
			}
		}
		break;
		
		default:
		hRes = E_UNUPPORTED_VALUE_TYPE;
		break;
	}
	
	if ( FAILED(hRes) )
	{
		return litex_error(hRes);
	}
	
	return val.Detach( pVal );
}


STDMETHODIMP CLiteRow::get_Count(LONG* pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	*pVal = m_nSize;
	return S_OK;
}

HRESULT CLiteRow::CreateObject( ILiteStatement* pStmt, IDispatch** ppRow)
{
	typedef CComObject<CLiteRow> _class;
	HRESULT hRes = E_POINTER;
	if (ppRow != NULL)
	{
		*ppRow = NULL;
		_class* p;
		hRes = _class::CreateInstance(&p);
		if (SUCCEEDED(hRes))
		{
			// If this object has ownership of the data then we need to keep it around
			hRes = p->Init(pStmt);
			if (SUCCEEDED(hRes))
			{
				hRes = p->_InternalQueryInterface( IID_IDispatch, (void**)ppRow);
			}
			
			if (FAILED(hRes))
			{
				delete p;
			}
		}
	}
	return hRes;
}

