/*
	Statement.cpp
*/

#include "stdafx.h"
#include "LiteX.h"
#include "Statement.h"
#include "Rows.h"
#include "Row.h"
#include "Columns.h"
#include "Parameters.h"
#include "Messages.h"
#include "Glob.h"

STDMETHODIMP CLiteStatement::InterfaceSupportsErrorInfo(REFIID riid)
{
	static const IID* arr[] = 
	{
		&IID_ILiteStatement
	};
	for (int i=0; i < sizeof(arr) / sizeof(arr[0]); i++)
	{
		if (IsEqualGUID(*arr[i],riid))
			return S_OK;
	}
	return S_FALSE;
}

HRESULT CLiteStatement::sqlite_error( sqlite3* db ) const
{
	HRESULT hr = sqlite_2_hr( sqlite3_errcode(db) );
#ifdef _UNICODE
	const WCHAR* pszDesc = (const WCHAR*)sqlite3_errmsg16(db);
#else
	const CHAR* pszDesc = (const CHAR*)sqlite3_errmsg(db);
#endif

	if ( FAILED(hr) )
	{
		return Error( pszDesc, IID_ILiteStatement, hr );
	}
	else
	{
		return hr;
	}
}

HRESULT CLiteStatement::sqlite_error( int nError ) const
{
	return litex_error( sqlite_2_hr( nError ) );
}

HRESULT CLiteStatement::litex_error( HRESULT hr ) const
{
	CComBSTR bsDesc;
	if ( FAILED(hr) && get_description( hr, bsDesc ) )
	{
		return Error( bsDesc, IID_ILiteStatement, hr );
	}
	else
	{
		return hr;
	}
}

STDMETHODIMP CLiteStatement::put_ActiveConnection(IDispatch *pVal)
{
	if ( m_eState == lxCreated )
	{
		return pVal->QueryInterface( IID_ILiteConnectionPrivate, (void**)&m_pConnection );
	}
	else
	{
		return litex_error( E_ALREADY_PREPARED );
	}
}

STDMETHODIMP CLiteStatement::get_ActiveConnection(IDispatch **pVal)
{
	return m_pConnection.QueryInterface( pVal );
}

STDMETHODIMP CLiteStatement::get_CommandText(BSTR *pVal)
{
	return m_bsCommand.CopyTo( pVal );
}

STDMETHODIMP CLiteStatement::put_CommandText(BSTR newVal)
{
	if ( m_eState == lxCreated )
	{
		m_bsCommand = newVal;
		return S_OK;
	}
	else
	{
		return litex_error( E_ALREADY_PREPARED );
	}
}

STDMETHODIMP CLiteStatement::Prepare( BSTR bsCommand )
{
#ifndef _UNICODE
	USES_CONV_UTF8;
#endif

	if ( m_eState != lxCreated )
	{
		return litex_error( E_ALREADY_PREPARED );
	}

	if ( !m_pConnection )
	{
		return litex_error( E_NO_ACTIVE_CONNECTION );
	}

	sqlite3* db;
	HRESULT hr = m_pConnection->GetDB( (void**)&db );
	if ( FAILED(hr) ) return litex_error( hr );

	if ( SysStringLen(bsCommand) > 0 )
	{
		m_bsCommand = bsCommand;
	}

	if ( m_bsCommand.Length() == 0 )
	{
		return litex_error( E_NO_COMMAND_STRING );
	}

	int nRes;
#ifdef _UNICODE
	nRes = sqlite3_prepare16_v2( db, m_bsCommand, -1, &m_stmt, NULL );
#else
	nRes = sqlite3_prepare_v2( db, W2UTF8(m_bsCommand), -1, &m_stmt, NULL );
#endif

	if ( nRes == SQLITE_OK )
	{
		m_eState = lxPrepared;
		m_bDone = false;
	}

	return sqlite_error( db );
}

static LxTypeEnum guess_type( const CComVariant& v )
{
	LxTypeEnum eType = lxUnknown;
	switch( V_VT(&v) & VT_TYPEMASK )
	{
		case VT_EMPTY:
		case VT_NULL:
		eType = lxNull;
		break;

		case VT_I1:
		case VT_UI2:
		case VT_I2:
		case VT_I4:
		case VT_INT:
		case VT_BOOL:
		case VT_UI4:
		case VT_UINT:
		eType = lxInteger;
		break;
		
		case VT_DECIMAL:
		{
			DECIMAL dm;
			if ( V_ISBYREF(&v) )
			{
				dm = *V_DECIMALREF(&v);
			}
			else
			{
				dm = V_DECIMAL(&v);
			}
			
			if ( is_decimal_integer( dm ) )
			{
				eType = lxInteger;
			}
			else
			{
				eType = lxFloat;
			}
		}
		break;

		case VT_R4:
		case VT_R8:
		case VT_CY:
		case VT_ERROR:
		eType = lxFloat;
		break;
		
		case VT_DATE:
		eType = lxDate;
		break;

		case VT_BSTR:
		case VT_LPSTR:
		case VT_LPWSTR:
		eType = lxString;
		break;

		case VT_UI1:
		if ( V_ISARRAY(&v) )
		{
			eType = lxBinary;
		}
		else
		{
			eType = lxInteger;
		}
		break;
	}

	return eType;
}

HRESULT CLiteStatement::bind_blob( CComPtr<IStream>& pSrc, LONG nColumn, int& nRes )
{
	HRESULT hr;
	STATSTG st;

	hr = pSrc->Stat( &st, STATFLAG_NONAME );
	if ( FAILED(hr) ) return hr;

	CComPtr<IStream> pStream;
	hr = CreateStreamOnHGlobal( NULL, TRUE, &pStream );
	if ( FAILED(hr) ) return hr;

	hr = pSrc->CopyTo( pStream, st.cbSize, NULL, NULL );
	if ( FAILED(hr) ) return hr;

	HGLOBAL hMem;
	hr = GetHGlobalFromStream( pStream, &hMem );
	if ( FAILED(hr) ) return hr;
	PVOID pMem = GlobalLock( hMem );
	nRes = sqlite3_bind_blob( m_stmt, nColumn, pMem, st.cbSize.LowPart, SQLITE_TRANSIENT );
	GlobalUnlock( hMem );
	return S_OK;
}

HRESULT CLiteStatement::bind_blob( SAFEARRAY* sa, LONG nColumn, int& nRes )
{
	HRESULT hr;
	PVOID pBuffer;

	hr = SafeArrayAccessData( sa, (void HUGEP**)&pBuffer );
	if ( FAILED(hr) ) return hr;
	nRes = sqlite3_bind_blob( m_stmt, nColumn, 
		pBuffer,
		sa->rgsabound[0].cElements,
		SQLITE_TRANSIENT );
	hr = SafeArrayUnaccessData( sa );
	return hr;
}


STDMETHODIMP CLiteStatement::BindParameter( VARIANT idx,  VARIANT vValue, LxTypeEnum eType )
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	int nParam;
	CComVariant v( vValue );
	HRESULT hr;

	hr = get_parameter_index( idx, nParam );
	if ( FAILED(hr) ) return hr;

	int nRes = SQLITE_OK;
	if ( eType == lxUnknown )
	{
		eType = guess_type( v );
	}

	switch( eType )
	{
		case lxNull:
		nRes = sqlite3_bind_null( m_stmt, nParam );
		break;

		case lxInteger:
		case lxLongInteger:
		{
			switch( V_VT(&v) & VT_TYPEMASK )
			{
				case VT_I1:
				case VT_UI2:
				case VT_I2:
				case VT_I4:
				case VT_INT:
				case VT_BOOL:
				hr = v.ChangeType( VT_INT );
				if ( FAILED(hr) ) return litex_error( hr );
				nRes = sqlite3_bind_int( m_stmt, nParam, V_INT(&v) );
				break;
				
				case VT_UI4:
				case VT_UINT:
				hr = v.ChangeType( VT_UINT );
				if ( FAILED(hr) ) return litex_error( hr );
				nRes = sqlite3_bind_int64( m_stmt, nParam, V_UINT(&v) );
				break;
				
				default:			
				hr = v.ChangeType( VT_DECIMAL );
				if ( FAILED(hr) ) return litex_error( hr );
				
				if ( is_decimal_integer( V_DECIMAL(&v) ) )
				{
					nRes = sqlite3_bind_int64( m_stmt, nParam, get_decimal_integer(V_DECIMAL(&v)) );
				}
				else
				{
					hr = v.ChangeType( VT_R8 );
					if ( FAILED(hr) ) return litex_error( hr );
					nRes = sqlite3_bind_double( m_stmt, nParam, V_R8(&v) );
				}
				break;
			}
		}
		break;
		
		case lxFloat:
		{
			hr = v.ChangeType( VT_R8 );
			if ( FAILED(hr) ) return litex_error( hr );
			nRes = sqlite3_bind_double( m_stmt, nParam, V_R8(&v) );
		}
		break;
		
		case lxDate:
		{
			hr = v.ChangeType( VT_DATE );
			if ( FAILED(hr) ) return litex_error( hr );
			nRes = sqlite3_bind_double( m_stmt, nParam, get_julian_date(V_DATE(&v)) );
		}
		break;

		case lxString:
		{
#ifndef _UNICODE
			USES_CONV_UTF8;
#endif
			hr = v.ChangeType( VT_BSTR );
			if ( FAILED(hr) ) return litex_error( hr );
#ifdef _UNICODE
			nRes = sqlite3_bind_text16( m_stmt, nParam, V_BSTR(&v), -1, SQLITE_TRANSIENT );
#else
			nRes = sqlite3_bind_text( m_stmt, nParam, W2UTF8( V_BSTR(&v) ), -1, SQLITE_TRANSIENT );
#endif
		}
		break;

		case lxBinary:
		{
			switch( V_VT(&v) & VT_TYPEMASK )
			{
				case VT_BSTR:
				case VT_LPSTR:
				case VT_LPWSTR:
				{
					hr = v.ChangeType( VT_BSTR );
					if ( FAILED(hr) ) return litex_error( hr );
					
					UINT nBytes = SysStringByteLen( V_BSTR(&v) );
					nRes = sqlite3_bind_blob( m_stmt, nParam, 
						V_BSTR(&v), nBytes,
						NULL );
				}
				break;

				case VT_UNKNOWN:
				{
					CComPtr<IStream> pStream;
					hr = V_UNKNOWN(&v)->QueryInterface( IID_IStream, (void**)&pStream );
					if ( hr == S_OK )
					{
						hr = bind_blob( pStream, nParam, nRes );
						if ( FAILED(hr) ) return litex_error( hr );
					}
					else
					{
						return litex_error( E_UNKNOWN_BINARY_DATA );
					}
				}
				break;

				case VT_UI1:
				if ( V_ISARRAY(&v) )
				{
					hr = bind_blob( v.parray, nParam, nRes );
				}
				else
				{
					return litex_error( E_UNKNOWN_BINARY_DATA );
				}
				break;

				default:
				return litex_error( E_UNKNOWN_BINARY_DATA );
			}
		}
		break;

		default:
		return litex_error( E_UNKNOWN_TYPE );
	}

	return sqlite_error( nRes );
}

STDMETHODIMP CLiteStatement::get_ColumnCount(LONG *pVal)
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}
	*pVal = sqlite3_column_count( m_stmt );
	return S_OK;
}

HRESULT CLiteStatement::get_column_name( LONG nColumn, CComBSTR& bsName ) const
{
#ifdef _UNICODE
	const WCHAR* pszName = (const WCHAR*)sqlite3_column_name16( m_stmt, nColumn );
	if ( pszName != NULL )
	{
		bsName = pszName;
	}
	else
	{
		return litex_error( E_NO_COLUMN_NAME );
	}
#else
	USES_CONV_UTF8;
	const CHAR* pszName = sqlite3_column_name( m_stmt, nColumn );
	if ( pszName != NULL )
	{
		bsName = UTF82W(pszName);
	}
	else
	{
		return litex_error( E_NO_COLUMN_NAME );
	}
#endif

	return S_OK;
}

STDMETHODIMP CLiteStatement::get_ColumnName(LONG nColumn, BSTR *pVal)
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	CComBSTR bs;
	HRESULT hr = get_column_name( nColumn, bs );
	if ( FAILED(hr) ) return hr;

	*pVal = bs.Detach();
	if ( *pVal != NULL )
	{
		return S_OK;
	}
	else
	{
		return litex_error( E_OUTOFMEMORY );
	}
}

LxTypeEnum CLiteStatement::get_column_type( int nColumn ) const
{
	LxTypeEnum eType;
	switch( sqlite3_column_type( m_stmt, nColumn ) )
	{
		case SQLITE_INTEGER: eType = lxInteger; break;
		case SQLITE_FLOAT: eType = lxFloat; break;
		case SQLITE_TEXT: eType = lxString; break;
		case SQLITE_BLOB: eType = lxBinary; break;
		case SQLITE_NULL: eType = lxNull; break;
		default: eType = lxUnknown; break;
	}
	return eType;
}

HRESULT CLiteStatement::get_column_index( const VARIANT& idx, int& nColumn ) const
{
	HRESULT hr;
	CComVariant v( idx );
	hr = variant_normalize( v );
	if ( FAILED(hr) ) return litex_error( hr );
	
	int nCount = sqlite3_column_count( m_stmt );
	switch( V_VT(&v) & VT_TYPEMASK )
	{
		case VT_INT: // integer
		nColumn = V_I4(&v);
		if ( nColumn < 0 || nColumn >= nCount )
		{
			hr = E_COLUMN_OUT_OF_RANGE;
		}
		break;

		case VT_BSTR: // column name
		{
			CComBSTR bsName;
			bool bFind = false;

			for( int i=0; i<nCount; i++ )
			{
				hr = get_column_name( i, bsName );
				if ( SUCCEEDED(hr) )
				{
					if ( VARCMP_EQ == compare_bstr( v.bstrVal, bsName ) )
					{
						nColumn = i;
						bFind = true;
						break;
					}
				}
			}

			if ( !bFind )
			{
				hr = E_UNKNOWN_COLUMN_NAME;
			}
		}
		break;

		default:
		ATLASSERT( FALSE );
		break;
	}

	if ( FAILED(hr) ) return litex_error( hr );
	return hr;
}

HRESULT CLiteStatement::get_parameter_index( const VARIANT& idx, int& nParameter ) const
{
	USES_CONV_UTF8;

	HRESULT hr;
	CComVariant v( idx );
	hr = variant_normalize( v );
	if ( FAILED(hr) ) return litex_error( hr );

	switch( V_VT(&v) & VT_TYPEMASK )
	{
		case VT_INT: // integer
		{
			int nCount = sqlite3_bind_parameter_count( m_stmt );
			if ( SUCCEEDED(hr) )
			{
				nParameter = V_I4(&v);
				if ( nParameter <= 0 || nParameter > nCount )
					hr = E_PARAM_OUT_OF_RANGE;
			}
		}
		break;

		case VT_BSTR: // column name
		{
			nParameter = sqlite3_bind_parameter_index( m_stmt, 
				W2UTF8( V_BSTR(&v) ) );

			if ( nParameter == 0 )
			{
				hr = E_UNKNOWN_PARAM_NAME;
			}
		}
		break;

		default: // shouldn't happend
		ATLASSERT( FALSE );
		break;
	}

	if ( FAILED(hr) ) return litex_error( hr );
	return hr;
}

STDMETHODIMP CLiteStatement::get_ColumnType( VARIANT idx, LxTypeEnum* eType )
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	int nColumn;
	HRESULT hr = get_column_index( idx, nColumn );
	if ( FAILED(hr) ) return hr;

	*eType = get_column_type( nColumn );

	return S_OK;
}

STDMETHODIMP CLiteStatement::get_ColumnValue( VARIANT idx, LxTypeEnum eType, VARIANT *pVal)
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	int nColumn;
	HRESULT hr = get_column_index( idx, nColumn );
	if ( FAILED(hr) ) return hr;

	if ( eType == lxUnknown )
	{
		eType = get_column_type( nColumn );
		if ( eType == lxUnknown ) return litex_error( E_UNKNOWN_COLUMN_TYPE );
	}

	CComVariant v;
	switch( eType )
	{
		case lxNull:
		V_VT(&v) = VT_NULL;
		break;

		case lxInteger:
		case lxLongInteger:
		{
			sqlite_int64 nVal = sqlite3_column_int64( m_stmt, nColumn );
			HRESULT hr = make_integer_or_decimal( v, nVal );
			if ( FAILED(hr) )
			{
				return litex_error( hr );
			}
		}
		break;

		case lxFloat:
		v = sqlite3_column_double( m_stmt, nColumn );
		break;
		
		case lxDate:
		V_VT(&v) = VT_DATE;
		V_DATE(&v) = gat_automation_date( sqlite3_column_double( m_stmt, nColumn ) );
		break;

		case lxString:
		{
#ifdef _UNICODE
			const WCHAR* pszValue = (const WCHAR*)sqlite3_column_text16( m_stmt, nColumn );
			if ( pszValue != NULL )
			{
				v = pszValue;
			}
			else
			{
				V_VT(&v) = VT_EMPTY;
			}
#else
			USES_CONV_UTF8;
			const CHAR* pszValue = (const CHAR*)sqlite3_column_text( m_stmt, nColumn );
			if ( pszValue != NULL )
			{
				v = UTF82W(pszValue);
			}
			else
			{
				V_VT(&v) = VT_EMPTY;
			}
#endif
		}
		break;

		case lxBinary:
		{
#ifdef _UNICODE
			int nColumnBytes = sqlite3_column_bytes16( m_stmt, nColumn );
#else
			int nColumnBytes = sqlite3_column_bytes( m_stmt, nColumn );
#endif
			v.parray = SafeArrayCreateVector( VT_UI1, 0, nColumnBytes );
			if ( v.parray )
			{
				PVOID pData;

				V_VT(&v) = VT_UI1|VT_ARRAY;
				hr = SafeArrayAccessData( v.parray, (void HUGEP**)&pData );
				if ( FAILED(hr) )
				{
					return litex_error( hr );
				}

				memcpy( pData, sqlite3_column_blob( m_stmt, nColumn ),
					nColumnBytes );

				hr = SafeArrayUnaccessData( v.parray );
				if ( FAILED(hr) )
				{
					return litex_error( hr );
				}
			}
			else
			{
				return litex_error( E_OUTOFMEMORY );
			}
		}
		break;
	}

	return v.Detach( pVal );
}

HRESULT CLiteStatement::create_row( LONG nSize, CComVariant& v )
{
	CComVariant val;
	HRESULT hr;

	v.parray = SafeArrayCreateVector( VT_VARIANT, 0, nSize );
	if ( v.parray )
	{
		V_VT(&v) = VT_VARIANT|VT_ARRAY;

		CComVariant idx;
		LxTypeEnum eType;
		for( LONG i=0; i<nSize; i++ )
		{
			idx = i;
			
			hr = get_ColumnType( idx, &eType );
			if ( FAILED(hr) )
			{
				return hr;
			}
			
			hr = get_ColumnValue( idx, eType, &val );
			if ( FAILED(hr) )
			{
				return hr;
			}

			hr = SafeArrayPutElement( v.parray, &i, &val );
			if ( FAILED(hr) )
			{
				return litex_error( hr );
			}
		}
		
		return hr;
	}
	else
	{
		return litex_error( E_OUTOFMEMORY );
	}
}

STDMETHODIMP CLiteStatement::get_Row( LxRowEnum eMode, VARIANT *pVal )
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	HRESULT hRes = S_OK;
	CComVariant v;
	switch( eMode )
	{
		case lxDefault:
		{
			LONG nColumns = sqlite3_column_count( m_stmt );
			switch( nColumns )
			{
				case 0: // NULL
				V_VT(&v) = VT_NULL;
				break;

				case 1: // no array, one scalar
				{
					CComVariant idx(0);
					hRes = get_ColumnValue( idx, lxUnknown, &v );
				}
				break;
				
				default:
				hRes = create_row( nColumns, v );
				break;
			}
		}
		break;
		
		case lxArray:
		{
			LONG nColumns = sqlite3_column_count( m_stmt );
			hRes = create_row( nColumns, v );
		}
		break;
		
		case lxCollection:
		{
			CComPtr<IDispatch> pRow;
			hRes = CLiteRow::CreateObject( this, &pRow );
			if ( SUCCEEDED(hRes) )
			{
				v = pRow;
			}
		}
		break;
		
		default:
		hRes = litex_error( E_OUTOFMEMORY );
		break;
	}
	
	if ( FAILED(hRes) )
	{
		return hRes;
	}
	else
	{
		return v.Detach( pVal );
	}
}

STDMETHODIMP CLiteStatement::Step( /*in,defaultvalue(1)*/ LONG nSteps, /*[out,retval]*/ VARIANT_BOOL *pbDone)
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	if ( nSteps <= 0 )
	{
		return litex_error( E_STMT_BAD_STEP_PARAMETER );
	}

	int nRes;
	bool bStop = false;

	while( (nSteps > 0) && !bStop )
	{
		nRes = sqlite3_step( m_stmt );
		switch( nRes )
		{
			case SQLITE_ROW: // OK
			break;

			case SQLITE_DONE: // OK but stop
			m_bDone = true;
			bStop = true;
			break;

			//case SQLITE_ERROR:
			//case SQLITE_MISUSE: // raise error
			default:
			{
				sqlite3* db;
				HRESULT hr = m_pConnection->GetDB( (void**)&db );
				if ( FAILED(hr) ) return litex_error( nRes );
				return sqlite_error( db );
			}
		}
		nSteps--;
	}

	*pbDone = (m_bDone? VARIANT_TRUE : VARIANT_FALSE );
	return sqlite_error( nRes );
}

STDMETHODIMP CLiteStatement::Execute()
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}
	
	int nRes = sqlite3_step( m_stmt );
	switch( nRes )
	{
		case SQLITE_ROW: // !!!
		return litex_error( E_STMT_NON_QUERY_RETURNS_ROW );

		case SQLITE_DONE: // OK but stop
		nRes = sqlite3_reset( m_stmt );
		if ( nRes != SQLITE_OK )
		{
			return sqlite_error( nRes );
		}
		break;

		//case SQLITE_ERROR:
		//case SQLITE_MISUSE: // raise error
		default:
		{
			sqlite3* db;
			HRESULT hr = m_pConnection->GetDB( (void**)&db );
			if ( FAILED(hr) ) return litex_error( nRes );
			return sqlite_error( db );
		}
	}
	
	return S_OK;
}

STDMETHODIMP CLiteStatement::get_Done( VARIANT_BOOL* pbDone )
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	*pbDone = (m_bDone? VARIANT_TRUE : VARIANT_FALSE );
	return S_OK;
}

STDMETHODIMP CLiteStatement::Close()
{
	if ( m_eState != lxPrepared )
	{
		return S_OK;
	}

	int nRes = sqlite3_finalize( m_stmt );
	if ( nRes == SQLITE_OK )
	{
		m_stmt = NULL;
		m_eState = lxCreated;
	}

	return sqlite_error( nRes );
}

STDMETHODIMP CLiteStatement::Reset()
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	int nRes = sqlite3_reset( m_stmt );
	if ( nRes == SQLITE_OK )
	{
		m_bDone = false;
	}

	return sqlite_error( nRes );
}

STDMETHODIMP CLiteStatement::get_State(LxStateEnum *pVal)
{
	*pVal = m_eState;
	return S_OK;
}

STDMETHODIMP CLiteStatement::get_ParameterCount(LONG *pVal)
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	*pVal = sqlite3_bind_parameter_count( m_stmt );
	return S_OK;
}

STDMETHODIMP CLiteStatement::get_ParameterName( LONG nIndex, BSTR *bsName )
{
	USES_CONV_UTF8;

	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	CComBSTR _bsName;
	const CHAR* pszName = sqlite3_bind_parameter_name( m_stmt, nIndex );
	if ( pszName != NULL )
	{
		_bsName = UTF82W(pszName);
	}
	else
	{
		_bsName = _T("");
	}

	return _bsName.CopyTo( bsName );
}

STDMETHODIMP CLiteStatement::get_RowCount( VARIANT *pVal )
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	// reset
	int nRes = sqlite3_reset( m_stmt );
	if ( nRes == SQLITE_OK )
	{
		m_bDone = false;
	}
	else
	{
		return sqlite_error( nRes );
	}

	// loop
	LONGLONG nCount = 0;
	bool bStop = false;

	while( !bStop )
	{
		nRes = sqlite3_step( m_stmt );
		switch( nRes )
		{
			case SQLITE_ROW: // OK
			nCount++;
			break;

			case SQLITE_DONE: // OK but stop
			bStop = true;
			break;

			//case SQLITE_ERROR:
			//case SQLITE_MISUSE: // raise error
			default:
			{
				sqlite3* db;
				HRESULT hr = m_pConnection->GetDB( (void**)&db );
				if ( FAILED(hr) ) return litex_error( nRes );
				return sqlite_error( db );
			}
		}
	}
	
	// reset
	nRes = sqlite3_reset( m_stmt );
	if ( nRes != SQLITE_OK )
	{
		return sqlite_error( nRes );
	}

	CComVariant v;
	HRESULT hr = make_integer_or_decimal( v, nCount );
	if ( FAILED(hr) )
	{
		return litex_error( hr );
	}
	return v.Detach( pVal );
}

STDMETHODIMP CLiteStatement::BindParameters( /*[in]*/ LPSAFEARRAY* params)
{
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}
	
	HRESULT hr = S_OK;
	LONG lLower;
	LONG lUpper;
	LONG p = 1;
	CComVariant v;
	CComVariant vIdx;
	VARTYPE vt = VT_EMPTY;

	hr = SafeArrayGetLBound( *params, 1, &lLower );
	if ( FAILED(hr) )
	{
		return litex_error( hr );
	}

	hr = SafeArrayGetUBound( *params, 1, &lUpper );
	if ( FAILED(hr) )
	{
		return litex_error( hr );
	}

	if ( lUpper >= lLower )
	{
		hr = SafeArrayGetVartype( *params, &vt );
		if ( FAILED(hr) )
		{
			return litex_error( hr );
		}
		ATLASSERT( vt == VT_VARIANT );
	}

	for( LONG i=lLower; i<=lUpper; i++,p++ )
	{
		hr = SafeArrayGetElement( *params, &i, &v );
		if ( FAILED(hr) )
		{
			return litex_error( hr );
		}
		
		vIdx = p;
		hr = BindParameter( vIdx, v, lxUnknown );
		if ( FAILED(hr) )
		{
			return litex_error( hr );
		}
	}

	return S_OK;	
}

STDMETHODIMP CLiteStatement::get_Rows(VARIANT_BOOL bStatic, LONG nMaxRecords, IDispatch **pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	
	if ( m_eState == lxCreated )
	{ // try to prepare
		CComBSTR bsEmpty;
		HRESULT hRes = Prepare( bsEmpty );
		if ( FAILED(hRes) )
		{
			return hRes;
		}
	}
	
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	HRESULT hRes = CLiteRows::CreateObject( !(bStatic == VARIANT_FALSE), nMaxRecords, this, pVal );
	return hRes;
}

STDMETHODIMP CLiteStatement::get_Columns(IDispatch **pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	
	if ( m_eState == lxCreated )
	{ // try to prepare
		CComBSTR bsEmpty;
		HRESULT hRes = Prepare( bsEmpty );
		if ( FAILED(hRes) )
		{
			return hRes;
		}
	}
	
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	HRESULT hRes = CLiteColumns::CreateObject( this, pVal );
	return hRes;
}

STDMETHODIMP CLiteStatement::get_Parameters(IDispatch** pVal)
{
	if ( pVal == NULL )
	{
		return E_POINTER;
	}
	
	if ( m_eState == lxCreated )
	{ // try to prepare
		CComBSTR bsEmpty;
		HRESULT hRes = Prepare( bsEmpty );
		if ( FAILED(hRes) )
		{
			return hRes;
		}
	}
	
	if ( m_eState != lxPrepared )
	{
		return litex_error( E_STMT_NOT_PREPARED );
	}

	HRESULT hRes = CLiteParameters::CreateObject( this, pVal );
	return hRes;
}
