/*
	LiteX.cpp
	
	This is LiteX++ library.
	LiteX++ is a simple C++ wrapper around SQLite3 C API.
	LiteX++ uses STL library.
	Most of class methods are inline and are included in header LiteX.hpp.
	Some additional functions are implemented in LiteX.cpp file.

	Define _LITEX_WITH_DELAYLOAD if you want delay-load stuff.
	Delay load support is only useful if:
		- LiteX++ is dynamically linked to sqlite3 API library - sqlite3.dll
		- sqlite3.dll library is delay loaded (/DELAYLOAD linker option)
		- LiteX Automation is registered in your system
	In these situations when your program wants to load sqlite3.dll
	LiteX Automation library is loaded (it may be UTF-16 version - sqlite3u.dll).
	So you don't need to copy sqlite3.dll library into system directory (very bad idea)
	or directory where your	application resides.
	Remeber that LiteX Automation library exports all native SQLite3 API functions and
	few own functions required by Automation.
	See documentation for more infromation.
	
	This code is public domain.
	You may use this code whenever you want.
	
	Author: Edmunt Pienkowsky
	E-mail: roed@onet.eu
*/

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#ifdef _LITEX_WITH_DELAYLOAD
#include <delayimp.h>
#endif
#include "LiteX.hpp"

static const char SQLITE_DLL[] = "sqlite3.dll";

namespace litex
{
	const TCHAR* SQLiteConnection::MEMORY_DB = _T(":memory:");

	std::wstring multi_byte_2_wide_char(
		const char* pStr,
		int nLen,
		UINT nCodePage )
	{
		assert( pStr != NULL );
		assert( nLen >= 0 || nLen == -1 );

		// figure out how many wide characters we are going to get 
		int nChars = MultiByteToWideChar( nCodePage , 0 , pStr , nLen , NULL , 0 ) ; 
		if ( nLen == -1 ) --nChars ; 
		if ( nChars == 0 ) return L"";

		// convert the narrow string to a wide string 
		// nb: slightly naughty to write directly into the string like this
		std::wstring buf;
		buf.resize( nChars ); 
		MultiByteToWideChar( nCodePage , 0 , pStr , nLen , 
			const_cast<wchar_t*>(buf.c_str()) , nChars ) ; 

		return buf;
	}

	std::string wide_char_2_multi_byte(
		const wchar_t* pStr , 
		int len,
		UINT nCodePage )
	{
		assert( pStr != NULL ) ; 
		assert( len >= 0 || len == -1 ); 

		// figure out how many narrow characters we are going to get 
		int nChars = WideCharToMultiByte( nCodePage , 0 , 
				pStr , len , NULL , 0 , NULL , NULL ) ; 
		if ( len == -1 ) --nChars; 
		if ( nChars == 0 ) return "";

		// convert the wide string to a narrow string
		// nb: slightly naughty to write directly into the string like this
		std::string buf ;
		buf.resize( nChars ) ;
		WideCharToMultiByte( nCodePage , 0 , pStr , len , 
			const_cast<char*>(buf.c_str()) , nChars , NULL , NULL ) ; 
		return buf ; 
	}

#ifdef _LITEX_WITH_DELAYLOAD
	namespace delayload
	{
		static bool find_litex_location( _tstring& sDll )
		{
			HKEY hKey;
			DWORD cbData= 0, dwType;
			LONG nRes = RegOpenKeyEx( HKEY_CLASSES_ROOT,
				_T("CLSID\\{3E22694D-7B92-42A1-89A7-668E2F7AA107}\\InprocServer32"), 
				0, KEY_READ, &hKey );
			if ( nRes != ERROR_SUCCESS )
			{
				return false;
			}

			nRes = RegQueryValueEx( hKey, NULL, 0, &dwType, NULL, &cbData );
			if ( nRes != ERROR_SUCCESS )
			{
				RegCloseKey( hKey );
				return false;
			}

			if ( dwType != REG_SZ )
			{
				RegCloseKey( hKey );
				return false;
			}

			BYTE* pBuffer = new BYTE[cbData];
			nRes = RegQueryValueEx( hKey, NULL, 0, &dwType, pBuffer, &cbData );

			if ( nRes == ERROR_SUCCESS )
			{
				sDll.assign( reinterpret_cast<LPCTSTR>(pBuffer), (cbData/sizeof(TCHAR)) - 1 );
				delete pBuffer;
				RegCloseKey( hKey );
				return true;
			}
			else
			{
				RegCloseKey( hKey );
				return false;
			}
		}

		static FARPROC WINAPI failureDelayHook(unsigned int dliNotify, PDelayLoadInfo pdli)
		{
			switch( dliNotify )
			{
				case dliFailLoadLib:
				if (_strcmpi( pdli->szDll, SQLITE_DLL ) == 0)
				{
					_tstring sDll;
					if ( find_litex_location( sDll ) )
					{
						return reinterpret_cast<FARPROC>(LoadLibrary(sDll.c_str()));
					}
				}
				break;

				case dliFailGetProc:
				default:
				return 0;
			}

			return 0;
		}

		static FARPROC WINAPI delayHook(unsigned int dliNotify, PDelayLoadInfo pdli)
		{
			switch (dliNotify)
			{
				case dliStartProcessing:
				// If you want to return control to the helper, return 0.
				// Otherwise, return a pointer to a FARPROC helper function
				// that will be used instead, thereby bypassing the rest 
				// of the helper.
				break;

				case dliNotePreLoadLibrary:
				// If you want to return control to the helper, return 0.
				// Otherwise, return your own HMODULE to be used by the 
				// helper instead of having it call LoadLibrary itself.
				if (_strcmpi( pdli->szDll, SQLITE_DLL ) == 0)
				{
					_tstring sDll;
					if ( find_litex_location( sDll ) )
					{
						return reinterpret_cast<FARPROC>(LoadLibrary(sDll.c_str()));
					}
				}
				break;

				case dliNotePreGetProcAddress:
				// If you want to return control to the helper, return 0.
				// If you choose you may supply your own FARPROC function 
				// address and bypass the helper's call to GetProcAddress.
				break;

				case dliFailLoadLib: 
				// LoadLibrary failed.
				// If you don't want to handle this failure yourself, return 0.
				// In this case the helper will raise an exception 
				// (ERROR_MOD_NOT_FOUND) and exit.
				// If you want to handle the failure by loading an alternate 
				// DLL (for example), then return the HMODULE for 
				// the alternate DLL. The helper will continue execution with 
				// this alternate DLL and attempt to find the
				// requested entrypoint via GetProcAddress.
				break;

				case dliFailGetProc:
				// GetProcAddress failed.
				// If you don't want to handle this failure yourself, return 0.
				// In this case the helper will raise an exception 
				// (ERROR_PROC_NOT_FOUND) and exit.
				// If you choose you may handle the failure by returning 
				// an alternate FARPROC function address.
				break;

				case dliNoteEndProcessing: 
				// This notification is called after all processing is done. 
				// There is no opportunity for modifying the helper's behavior
				// at this point except by longjmp()/throw()/RaiseException. 
				// No return value is processed.
				break;

				default:
				return NULL;
			}

			return NULL;
		}

		void set_handler() throw()
		{
#if _MSC_VER <= 1200
			__pfnDliNotifyHook = delayHook;
			//__pfnDliFailureHook = failureDelayHook;
#else
			__pfnDliNotifyHook2 = delayHook;
			//__pfnDliFailureHook2 = failureDelayHook;
#endif
		}

		static int exception_filter(unsigned int code, struct _EXCEPTION_POINTERS *ep)
		{
			if (code == VcppException(ERROR_SEVERITY_ERROR,ERROR_MOD_NOT_FOUND))
			{
				PDelayLoadInfo pdli = 
					reinterpret_cast<PDelayLoadInfo>(ep->ExceptionRecord->ExceptionInformation[0]);
				if (_strcmpi( pdli->szDll, SQLITE_DLL ) == 0)
				{
					return EXCEPTION_EXECUTE_HANDLER;
				}
				else
				{
					return EXCEPTION_CONTINUE_SEARCH;
				}
			}
			else
			{
				return EXCEPTION_CONTINUE_SEARCH;
			}
		};


		HRESULT load_library( bool bAutomation )
		{
			if ( bAutomation )
			{
				set_handler();
			}

			__try
			{
#if _MSC_VER > 1200
				return __HrLoadAllImportsForDll( SQLITE_DLL );
#else
				return sqlite3_libversion_number(),S_OK;
#endif
			}
			__except( exception_filter( GetExceptionCode(), GetExceptionInformation() ) )
			{
				return HRESULT_FROM_WIN32( ERROR_MOD_NOT_FOUND );
			}
		}

		bool free_library()
		{
#if _MSC_VER <= 1200
			return (__FUnloadDelayLoadedDLL( SQLITE_DLL ) != 0);
#else
			return (__FUnloadDelayLoadedDLL2( SQLITE_DLL ) != 0);
#endif
		}
	};
#endif
};
