/*
	LiteX.hpp
	
	This is LiteX++ library.
	LiteX++ is a simple C++ wrapper around SQLite3 C API.
	LiteX++ uses STL library.
	Most of class methods are inline and are included in header LiteX.hpp.
	Some additional functions are implemented in LiteX.cpp file.

	Define _LITEX_WITH_DELAYLOAD if you want delay-load stuff.
	Delay load support is only useful if:
		- LiteX++ is dynamically linked to sqlite3 API library - sqlite3.dll
		- sqlite3.dll library is delay loaded (/DELAYLOAD linker option)
		- LiteX Automation is registered in your system
	In these situations when your program wants to load sqlite3.dll
	LiteX Automation library is loaded (it may be UTF-16 version - sqlite3u.dll).
	So you don't need to copy sqlite3.dll library into system directory (very bad idea)
	or directory where your	application resides.
	Remeber that LiteX Automation library exports all native SQLite3 API functions and
	few own functions required by Automation.
	See documentation for more infromation.
	
	This code is public domain.
	You may use this code whenever you want.
	
	Author: Edmunt Pienkowsky
	E-mail: roed@onet.eu
*/

#ifndef __cplusplus
#error This is C++ wrapper for SQLite3 library.
#error Use C native library in C project.
#endif

#ifndef _LITEX_HPP_
#define _LITEX_HPP_

#include <sqlite3.h>
#include <tchar.h>
#include <string>
#include <cassert>
#include <stdexcept>


// all libray stuff included into litex namespace
namespace litex
{
#ifdef _LITEX_WITH_DELAYLOAD
	namespace delayload
	{
		/*
			Installs own delay-load DLL notification hook.
			When sqlite3.dll library is needed it tries to
			locate LiteX Automation library.
			If LiteX Automation library cannot be found standard
			search procedure will be used.

			Note that LiteX Automation libray may be Unicode
			version - sqlite3u.dll.
		*/
		void set_handler();

		/*
			Loads sqlite3.dll library.
			If bAutomation is true set_handler() is called before.

			You may call this routine before any call to SQLite engine
			to make sure that this engine is accessible.
			S_OK (0) - success
			0x8007007e is the most common error code.

			Warning: Don't call this routine from DllMain function!
		*/
		HRESULT load_library( bool bAutomation = true );

		/*
			Unloads previously loaded sqlite3(u).dll library.
			You do not need to call this routine.
		*/
		bool free_library();
	}
#endif

	// _tstring - our string class

#ifdef _UNICODE
	typedef std::wstring _tstring;
#else
	typedef std::string _tstring;
#endif

	// string encodeing/decoding routines 

	std::wstring multi_byte_2_wide_char(
		const char* pStr,
		int nLen = -1,
		UINT nCodePage = CP_ACP );

	std::string wide_char_2_multi_byte(
		const wchar_t* pStr , 
		int len = -1,
		UINT nCodePage = CP_ACP );

	inline std::string utf8_2_ansi( const char* psz )
	{
		std::wstring ws = multi_byte_2_wide_char( psz, -1, CP_UTF8 );
		return wide_char_2_multi_byte( ws.c_str(), (int)ws.length(), CP_ACP );
	}

	inline std::wstring utf8_2_utf16( const char* psz )
	{
		return multi_byte_2_wide_char( psz, -1, CP_UTF8 );
	}

	inline std::string ansi_2_utf8( const char* psz )
	{
		std::wstring ws = multi_byte_2_wide_char( psz, -1, CP_ACP );
		return wide_char_2_multi_byte( ws.c_str(), (int)ws.length(), CP_UTF8 );
	}

	inline std::string utf16_2_utf8( const wchar_t* psz )
	{
		return wide_char_2_multi_byte( psz, -1, CP_UTF8 );
	}
	
	inline std::string utf16_2_ansi( const wchar_t* psz )
	{
		return wide_char_2_multi_byte( psz, -1, CP_ACP );
	}
	
	inline std::wstring ansi_2_utf16( const char* psz )
	{
		return multi_byte_2_wide_char( psz, -1, CP_ACP );
	}
	
	/*
		LiteX++ library runtime exception.
	*/
	class SQLiteRuntimeException :public std::runtime_error
	{
		public:
		SQLiteRuntimeException( const _tstring& sMsg )
#ifdef _UNICODE
			:runtime_error(utf16_2_ansi(sMsg.c_str()))
#else
			:runtime_error(sMsg.c_str())
#endif
		{}
		
		SQLiteRuntimeException( const SQLiteRuntimeException& ex )
			:std::runtime_error( ex.what() )
		{}
		
		_tstring get_Message() const
		{
#ifdef _UNICODE
			return ansi_2_utf16( what() );	
#else
			return what();
#endif	
		}
		
		static void Throw( const _tstring& sMsg )
		{
			throw SQLiteRuntimeException( sMsg );
		}
		
	};

	/*
		SQLite3 exception - error from SQLite3 library.
	*/
	class SQLiteException :public std::runtime_error
	{
		public:

		SQLiteException( sqlite3* pDb ) :runtime_error("")
		{
			assert( pDb != NULL );

			m_nErrorCode = sqlite3_errcode( pDb );
#ifdef _UNICODE
			m_sMessage = static_cast<const wchar_t*>( sqlite3_errmsg16(pDb) );
			m_sAnsiMessage = utf16_2_ansi( m_sMessage.c_str() );
#else
			m_sAnsiMessage = utf8_2_ansi( sqlite3_errmsg(pDb) );
#endif
		}

		SQLiteException( sqlite3* pDb, int nErrorCode )
			:runtime_error(""),m_nErrorCode(nErrorCode)
		{
			//assert( pDb != NULL );
#ifdef _UNICODE
			m_sMessage = static_cast<const wchar_t*>( sqlite3_errmsg16(pDb) );
			m_sAnsiMessage = utf16_2_ansi( m_sMessage.c_str() );
#else
			m_sAnsiMessage = utf8_2_ansi( sqlite3_errmsg(pDb) );
#endif
		}
		
		SQLiteException( int nErrorCode, const char* pszMsg )
			:runtime_error(""),m_nErrorCode(nErrorCode)
		{
#ifdef _UNICODE
			m_sMessage = utf8_2_utf16( pszMsg );
			m_sAnsiMessage = utf8_2_ansi( pszMsg );
#else
			m_sAnsiMessage = utf8_2_ansi( pszMsg );
#endif
		}

		SQLiteException( const SQLiteException& exp ) 
			:runtime_error(exp.m_sAnsiMessage),m_sAnsiMessage(exp.m_sAnsiMessage),m_nErrorCode(exp.m_nErrorCode)
#ifdef _UNICODE		
			,m_sMessage(exp.m_sMessage)
#endif
		{
		}

		virtual ~SQLiteException() {}

		static void Throw( int nErrorCode, sqlite3* pDb )
		{
			if ( nErrorCode != SQLITE_OK )
			{
				throw SQLiteException( pDb, nErrorCode );
			}
		}

		static void Throw( int nErrorCode )
		{
			if ( nErrorCode != SQLITE_OK )
			{
				throw SQLiteException( NULL, nErrorCode );
			}
		}
		
		static void Throw( int nErrorCode, const char* pszMsg )
		{
			if ( nErrorCode != SQLITE_OK )
			{
				throw SQLiteException( nErrorCode, pszMsg );
			}
		}

		public:

		int get_ErrorCode() const
		{
			return m_nErrorCode;
		}

		const _tstring& get_Message() const
		{
#ifdef _UNICODE		
			return m_sMessage;
#else
			return m_sAnsiMessage;
#endif
		}
		
		virtual const char *what() const
		{
			return m_sAnsiMessage.c_str();
		}
		
		protected:

		int m_nErrorCode;
#ifdef _UNICODE		
		_tstring m_sMessage;
#endif
		std::string m_sAnsiMessage;

	};
	
	// forward declaration
	class SQLiteStatement;

	/*
		Wrapper around sqlite3*
	*/
	class SQLiteConnection
	{
		public:

		// version string
		static _tstring get_VersionString()
		{
#ifdef _UNICODE
			return utf8_2_utf16( sqlite3_libversion() );
#else
			return utf8_2_ansi( sqlite3_libversion() );
#endif
		}

		// version number
		static int get_VersionNumber()
		{
			return sqlite3_libversion_number();
		}

		// in-memory database name
		static const TCHAR* MEMORY_DB;

		public:

		SQLiteConnection()
			:m_pDb(NULL)
		{
		}

		SQLiteConnection( const _tstring& sDbPath )
			:m_pDb(NULL)
		{
			_open( sDbPath.c_str(), true );
		}
		
		SQLiteConnection( const TCHAR* pszDbPath )
			:m_pDb(NULL)
		{
			_open( pszDbPath, true );
		}
		
		SQLiteConnection( SQLiteConnection& db )
			:m_pDb(db._detach())
		{
		}
		
		SQLiteConnection& operator=( SQLiteConnection& db )
		{
			m_pDb = db._detach();
			return *this;
		}

		virtual ~SQLiteConnection()
		{
			if ( m_pDb != NULL )
			{
				sqlite3_interrupt( m_pDb );
				sqlite3_close( m_pDb );
			}
		}

		/*
			Opens database.
			This method doesn't raise exception.
		*/
		bool Open( const _tstring& sDbPath )
		{
			assert( m_pDb == NULL );
			return (_open( sDbPath.c_str(), false ) == SQLITE_OK);
		}

		/*
			Opens in-memory database.
			This method doesn't raise exception.
		*/
		bool OpenInMemory()
		{
			assert( m_pDb == NULL );
			return (_open( MEMORY_DB, false ) == SQLITE_OK);
		}

		void Close()
		{
			if ( m_pDb != NULL )
			{
				SQLiteException::Throw( 
					sqlite3_close( m_pDb ),
					m_pDb );
				m_pDb = NULL;
			}
		}

		void Interrupt()
		{
			assert( m_pDb != NULL );
			sqlite3_interrupt( m_pDb );
		}

		/*
			Access to sqlite* handle.
		*/
		operator sqlite3*() const
		{
			assert( m_pDb != NULL );
			return const_cast<sqlite3*>(m_pDb);
		}
		
		operator bool() const
		{
			return (m_pDb != NULL);
		}

		sqlite_int64 get_LastInsertRowid() const
		{
			assert( m_pDb != NULL );
			return sqlite3_last_insert_rowid( const_cast<sqlite3*>(m_pDb) );
		}

		int get_Changes() const
		{
			assert( m_pDb != NULL );
			return sqlite3_changes( const_cast<sqlite3*>(m_pDb) );
		}

		int get_TotalChanges() const
		{
			assert( m_pDb != NULL );
			return sqlite3_total_changes( const_cast<sqlite3*>(m_pDb) );
		}

		void set_BusyTimeout( int nMs )
		{
			assert( m_pDb != NULL );
			sqlite3_busy_timeout( m_pDb, nMs );
		}

		/*
			Executing many commands at once.
		*/
		void BatchExecute( const TCHAR* pszSql )
		{
			assert( m_pDb != NULL );
			assert( pszSql != NULL );
			
			char* pszMsg = NULL;
			int nRes = SQLITE_OK;
			
#ifdef _UNICODE			
			nRes = sqlite3_exec( m_pDb, utf16_2_utf8(pszSql).c_str(), NULL, NULL, &pszMsg );
#else
			nRes = sqlite3_exec( m_pDb, ansi_2_utf8(pszSql).c_str(), NULL, NULL, &pszMsg );
#endif
			std::string sMsg;
			if ( pszMsg != NULL )
			{
				sMsg = pszMsg;
				sqlite3_free( pszMsg );
			}
			SQLiteException::Throw( nRes, sMsg.c_str() );
		}
		
		void BatchExecute( const _tstring& sSql )
		{
			BatchExecute( sSql.c_str() );
		}
		
		/*
			Executes single SQL tatement without result(s).
		*/
		inline void ExecuteNonQuery( const TCHAR* szSql );
		void ExecuteNonQuery( const _tstring& sSql )
		{
			ExecuteNonQuery( sSql.c_str() );
		}
		
		/*
			Executes singleton SQL statement.
		*/
		inline int ExecuteScalarInt( const _tstring& sSql );
		inline sqlite_int64 ExecuteScalarInt64( const _tstring& sSql );
		inline double ExecuteScalarDouble( const _tstring& sSql );
		inline _tstring ExecuteScalarText( const _tstring& sSql );
		
		inline int ExecuteScalarInt( const TCHAR* pszSql );
		inline sqlite_int64 ExecuteScalarInt64( const TCHAR* pszSql );
		inline double ExecuteScalarDouble( const TCHAR* pszSql );
		inline _tstring ExecuteScalarText( const TCHAR* pszSql );
		
		/*
			Prepares single SQL statement.
		*/
		inline SQLiteStatement Prepare( const _tstring& sSql );
		inline SQLiteStatement Prepare( const TCHAR* pszSql );
		
		/*
			Transaction shortcuts.
		*/
		void BeginTransaction()
		{
			ExecuteNonQuery(_T("BEGIN TRANSACTION"));
		}
		
		void CommitTransaction()
		{
			ExecuteNonQuery(_T("COMMIT TRANSACTION"));
		}
		
		void RollbackTransaction()
		{
			ExecuteNonQuery(_T("ROLLBACK TRANSACTION"));
		}

		protected:

		int _open( const TCHAR* pszDbPath, bool bThrowException )
		{
			assert( m_pDb == NULL );
			assert( pszDbPath != NULL );

			int nRes = SQLITE_OK;
#ifdef _UNICODE
			nRes = sqlite3_open16( static_cast<const void*>(pszDbPath), &m_pDb );
#else
			nRes = sqlite3_open( ansi_2_utf8(pszDbPath).c_str(), &m_pDb );
#endif
			if ( bThrowException )
			{
				SQLiteException::Throw( nRes );
			}
			return nRes;
		}
		
		sqlite3* _detach()
		{
			sqlite3* pDb = m_pDb;
			m_pDb = NULL;
			return pDb;
		}

		protected:

		sqlite3* m_pDb;
	};

	/*
		Wrapper around sqlite3_stmt*.
	*/
	class SQLiteStatement
	{
		public:

		SQLiteStatement( SQLiteConnection& connection )
			:m_connection(connection),m_pStmt(NULL)
		{
		}
		
		SQLiteStatement( SQLiteConnection& connection, const TCHAR* pszSql )
			:m_connection(connection),m_pStmt(NULL)
		{
			_prepare( pszSql );
		}		

		SQLiteStatement( SQLiteConnection& connection, const _tstring& sSql )
			:m_connection(connection),m_pStmt(NULL)
		{
			_prepare( sSql.c_str() );
		}

		SQLiteStatement( SQLiteStatement& stmt )
			:m_connection(stmt.m_connection),m_pStmt(stmt._detach())
		{
		}

		SQLiteStatement& operator=( SQLiteStatement& stmt )
		{
			m_connection = stmt.m_connection;
			m_pStmt = stmt._detach();
			return *this;
		}

		virtual ~SQLiteStatement()
		{
			if ( m_pStmt )
			{
				sqlite3_finalize( m_pStmt );
			}
		}

		/*
			Access to sqlite3_stmt* handle.
		*/
		operator sqlite3_stmt*() const
		{
			return const_cast<sqlite3_stmt*>(m_pStmt);
		}
		
		operator bool() const
		{
			return (m_pStmt != NULL);
		}

		SQLiteConnection& get_Connection() const
		{
			return const_cast<SQLiteConnection&>(m_connection);
		}
		
		void Prepare( const _tstring& sSql )
		{
			_prepare( sSql.c_str() );
		}
		
		void Prepare( const TCHAR* pszSql )
		{
			_prepare( pszSql );
		}

		void Reset()
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_reset( m_pStmt ),
				m_connection );
		}

		void Finalize()
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_finalize( m_pStmt ),
				m_connection );
			m_pStmt = NULL;
		}

		int get_ParameterCount() const
		{
			assert( m_pStmt != NULL );
			return sqlite3_bind_parameter_count( const_cast<sqlite3_stmt*>(m_pStmt) );
		}

		_tstring get_ParameterName( int nParam ) const
		{
			assert( m_pStmt != NULL );
#ifdef _UNICODE
			return utf8_2_utf16( sqlite3_bind_parameter_name( m_pStmt, nParam ) );
#else
			return utf8_2_ansi( sqlite3_bind_parameter_name( m_pStmt, nParam ) );
#endif
		}

		/*
			Data binding.
		*/
		void BindBlob( int nParam, const void* pBlock, int nSize )
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_bind_blob( m_pStmt, nParam, pBlock, nSize, SQLITE_STATIC ),
				m_connection );
		}

		void BindBlob( const _tstring& sParam, const void* pBlock, int nSize )
		{
			BindBlob( _parameter_index(sParam.c_str()), pBlock, nSize );
		}
		
		void BindBlob( const TCHAR* pszParam, const void* pBlock, int nSize )
		{
			BindBlob( _parameter_index(pszParam), pBlock, nSize );
		}
		
		void BindDouble( int nParam, double val )
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_bind_double( m_pStmt, nParam, val ),
				m_connection );
		}

		void BindDouble( const _tstring& sParam, double val )
		{
			BindDouble( _parameter_index(sParam.c_str()), val );
		}
		
		void BindDouble( const TCHAR* pszParam, double val )
		{
			BindDouble( _parameter_index(pszParam), val );
		}

		void BindInt( int nParam, int val )
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_bind_int( m_pStmt, nParam, val ),
				m_connection );
		}

		void BindInt( const _tstring& sParam, int val )
		{
			BindInt( _parameter_index(sParam.c_str()), val );
		}
		
		void BindInt( const TCHAR* pszParam, int val )
		{
			BindInt( _parameter_index(pszParam), val );
		}

		void BindInt64( int nParam, sqlite_int64 val )
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_bind_int64( m_pStmt, nParam, val ),
				m_connection );
		}

		void BindInt64( const _tstring& sParam, sqlite_int64 val )
		{
			BindInt64( _parameter_index(sParam.c_str()), val );
		}
		
		void BindInt64( const TCHAR* pszParam, sqlite_int64 val )
		{
			BindInt64( _parameter_index(pszParam), val );
		}

		void BindText( int nParam, const TCHAR* val )
		{
			assert( m_pStmt != NULL );
			assert( val != NULL );
			
#ifdef _UNICODE
			SQLiteException::Throw(
				sqlite3_bind_text16( m_pStmt, nParam, 
					static_cast<const void*>(val), -1, SQLITE_TRANSIENT ),
				m_connection );				
#else
			SQLiteException::Throw(
				sqlite3_bind_text( m_pStmt, nParam, 
					ansi_2_utf8(val).c_str(), -1, SQLITE_TRANSIENT ),
				m_connection );				
#endif
		}
		
		void BindText( int nParam, const _tstring& val )
		{
			BindText( nParam, val.c_str() );
		}
		
		void BindText( const _tstring& sParam, const TCHAR* val )
		{
			BindText( _parameter_index(sParam.c_str()), val );
		}

		void BindText( const _tstring& sParam, const _tstring& val )
		{
			BindText( _parameter_index(sParam.c_str()), val.c_str() );
		}

		void BindText( const TCHAR* pszParam, const TCHAR* val )
		{
			BindText( _parameter_index(pszParam), val );
		}
		
		void BindText( const TCHAR* pszParam, const _tstring& val )
		{
			BindText( _parameter_index(pszParam), val.c_str() );
		}
		
		void BindNull( int nParam )
		{
			assert( m_pStmt != NULL );
			SQLiteException::Throw( 
				sqlite3_bind_null( m_pStmt, nParam ),
				m_connection );
		}

		void BindNull( const _tstring& sParam )
		{
			BindNull( _parameter_index(sParam.c_str()) );
		}
		
		void BindNull( const TCHAR* pszParam )
		{
			BindNull( _parameter_index(pszParam) );
		}

		// =====================================================
		
		/*
			Column access.
		*/

		int get_ColumnCount() const
		{
			assert( m_pStmt != NULL );
			return sqlite3_column_count( const_cast<sqlite3_stmt*>(m_pStmt) );
		}

		_tstring get_ColumnName( int nColIdx )
		{
			assert( m_pStmt != NULL );
#ifdef _UNICODE
			return static_cast<const wchar_t*>(sqlite3_column_name16(m_pStmt,nColIdx));
#else
			return utf8_2_ansi( sqlite3_column_name(m_pStmt,nColIdx) );
#endif
		}

		_tstring get_ColumnDecltype( int nColIdx )
		{
			assert( m_pStmt != NULL );
#ifdef _UNICODE
			return static_cast<const wchar_t*>(sqlite3_column_decltype16(m_pStmt,nColIdx));
#else
			return utf8_2_ansi( sqlite3_column_decltype(m_pStmt,nColIdx) );
#endif
		}

		bool Step()
		{
			assert( m_pStmt != NULL );
			int nRes = sqlite3_step( m_pStmt );
			switch ( nRes )
			{
				case SQLITE_DONE:
				return false;

				case SQLITE_ROW:
				return true;

				default:
				SQLiteException::Throw( nRes, m_connection );
				return false;
			}
		}
		
		void Execute()
		{
			assert( m_pStmt != NULL );
			int nRes = sqlite3_step( m_pStmt );
			switch ( nRes )
			{
				case SQLITE_DONE:
				SQLiteException::Throw( sqlite3_reset( m_pStmt ), m_connection );
				break;

				case SQLITE_ROW:
				SQLiteRuntimeException::Throw( _T("Non-query statement returns row.") );				

				default:
				SQLiteException::Throw( nRes, m_connection );
			}
		}
		

		int get_DataCount() const
		{
			assert( m_pStmt != NULL );
			return sqlite3_data_count( const_cast<sqlite3_stmt*>(m_pStmt) );
		}

		const void* get_ColumnBlob( int nColIdx, int& nBlobSize ) const
		{
			assert( m_pStmt != NULL );
			nBlobSize = sqlite3_column_bytes( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx );
			return sqlite3_column_blob( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx );
		}

		double get_ColumnDouble( int nColIdx ) const
		{
			assert( m_pStmt != NULL );
			return sqlite3_column_double( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx );
		}

		int get_ColumnInt( int nColIdx ) const
		{
			assert( m_pStmt != NULL );
			return sqlite3_column_int( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx );
		}

		sqlite_int64 get_ColumnInt64( int nColIdx ) const
		{
			assert( m_pStmt != NULL );
			return sqlite3_column_int64( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx );
		}

		_tstring get_ColumnText( int nColIdx ) const
		{
#ifdef _UNICODE
			return static_cast<const wchar_t*>(sqlite3_column_text16( m_pStmt, nColIdx ));
#else
			return utf8_2_ansi( reinterpret_cast<const char*>(sqlite3_column_text( m_pStmt, nColIdx )) );
#endif
		}

		int get_ColumnType( int nColIdx ) const
		{
			assert( m_pStmt != NULL );
			return sqlite3_column_type( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx );
		}

		bool IsNull( int nColIdx ) const
		{
			assert( m_pStmt != NULL );
			return (SQLITE_NULL == sqlite3_column_type( const_cast<sqlite3_stmt*>(m_pStmt), nColIdx ));
		}

		protected:

		int _parameter_index( const TCHAR* pszParam )
		{
			assert( pszParam != NULL );
			
#ifdef _UNICODE
			return sqlite3_bind_parameter_index( 
				m_pStmt, utf16_2_utf8(pszParam).c_str() );
#else
			return sqlite3_bind_parameter_index(
				m_pStmt, ansi_2_utf8(pszParam).c_str() );
#endif
		}

		sqlite3_stmt* _detach()
		{
			sqlite3_stmt* pStmt = m_pStmt;
			m_pStmt = NULL;
			return pStmt;
		}
		
		protected:
		
		void _prepare( const TCHAR* pszStatement )
		{
			assert( m_pStmt == NULL );
			assert( pszStatement != NULL );

#ifdef _UNICODE
			SQLiteException::Throw( sqlite3_prepare16_v2( 
					m_connection,
					static_cast<const void*>(pszStatement),
					-1,
					&m_pStmt, NULL ),
				m_connection );
#else
			SQLiteException::Throw( sqlite3_prepare_v2( 
					m_connection,
					ansi_2_utf8(pszStatement).c_str(),
					-1,
					&m_pStmt, NULL ),
				m_connection );
#endif
		}

		protected:

		SQLiteConnection& m_connection;
		sqlite3_stmt* m_pStmt;
	};
	
	inline void SQLiteConnection::ExecuteNonQuery(const TCHAR* szSql)
	{
		SQLiteStatement stmt( *this, szSql );
		stmt.Step();
	}
	
	inline int SQLiteConnection::ExecuteScalarInt( const _tstring& sSql )
	{
		SQLiteStatement stmt( *this, sSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnInt(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarInt: Query doesn't return any results.") );
			return 0;
		}
	}
	
	inline int SQLiteConnection::ExecuteScalarInt( const TCHAR* pszSql )
	{
		SQLiteStatement stmt( *this, pszSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnInt(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarInt: Query doesn't return any results.") );
			return 0;
		}
	}
	
	inline sqlite_int64 SQLiteConnection::ExecuteScalarInt64( const _tstring& sSql )
	{
		SQLiteStatement stmt( *this, sSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnInt64(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarInt64: Query doesn't return any results.") );		
			return 0;
		}
	}
	
	inline sqlite_int64 SQLiteConnection::ExecuteScalarInt64( const TCHAR* pszSql )
	{
		SQLiteStatement stmt( *this, pszSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnInt64(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarInt64: Query doesn't return any results.") );		
			return 0;
		}
	}
	
	inline double SQLiteConnection::ExecuteScalarDouble( const _tstring& sSql )
	{
		SQLiteStatement stmt( *this, sSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnDouble(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarDouble: Query doesn't return any results.") );		
			return 0.0;
		}
	}
	
	inline double SQLiteConnection::ExecuteScalarDouble( const TCHAR* pszSql )
	{
		SQLiteStatement stmt( *this, pszSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnDouble(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarDouble: Query doesn't return any results.") );		
			return 0.0;
		}
	}
	
	inline _tstring SQLiteConnection::ExecuteScalarText( const _tstring& sSql )
	{
		SQLiteStatement stmt( *this, sSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnText(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarText: Query doesn't return any results.") );		
			return _T("");
		}
	}
	
	inline _tstring SQLiteConnection::ExecuteScalarText( const TCHAR* pszSql )
	{
		SQLiteStatement stmt( *this, pszSql );
		if ( stmt.Step() )
		{
			return stmt.get_ColumnText(0);
		}
		else
		{
			SQLiteRuntimeException::Throw( _T("ExecuteScalarText: Query doesn't return any results.") );		
			return _T("");
		}
	}

	inline SQLiteStatement SQLiteConnection::Prepare( const _tstring& sSql )
	{
		return SQLiteStatement( *this, sSql );
	}
	
	inline SQLiteStatement SQLiteConnection::Prepare( const TCHAR* pszSql )
	{
		return SQLiteStatement( *this, pszSql );
	}

};

#endif
